/*
   A simple timer to measure the elapsed time,
   activated by two capcitive sensors.

   You may use, distribute and modify this code under the
   terms of the GNU license
*/

#include <CapacitiveSensor.h>
#include <LiquidCrystal.h>

/*
   Constant defining the amount of time of a single slot during the prepare time.
   The total prepare time is 16 x PREPARE_INTERVAL_SLOT
*/
const int PREPARE_INTERVAL_SLOT = 120;

/*
   Firmware version to display at the beginning
*/
const String FIRMWARE_VERSION = "1.0.4";

/*
   The threshold above which the sensor is touched
*/
const int CAPACITY_SENSOR_THRESHOLD = 200;

/*
   New LiquidCrystal object
*/
LiquidCrystal lcd(12, 11, 9, 8, 7, 6);

/*
   Define 2 capacity sensor, one for each hand.
*/
CapacitiveSensor leftHandCapacitySensor  = CapacitiveSensor(3, 2);
CapacitiveSensor rightHandCapacitySensor = CapacitiveSensor(5, 4);

/*
   A couple of variables to store:
   - start time (when the timer is running)
   - prepare time (when the timer is preparing to run)
   - last and best elapsed times
*/
long startTime = 0,
     prepareTime = 0,
     lastTime = 0,
     bestTime = 0;

/*
   Firmware version string
*/
String versionString = "Firmware v.";

/*
   Set of the different modes the timer is running on
*/
enum {
  MODE_INIT,     // at the very beginning
  MODE_IDLE,     // when there's nothing to do
  MODE_PREPARE,  // preparing to measure the time
  MODE_READY,    // ready to start
  MODE_RUNNING,  // running the timer
  MODE_STOP      // timer stopped
} modes;

/*
   A variable that holds the current mode
*/
int currentMode = MODE_INIT;  // at the beginning, starting in init mode

/*
    During the prepare phase, a pad character (all leds on) is printed
    on the 2nd row of the display.
    This variable holds the position where the carachter is printed.
*/
int prepareCursor;

/*
   Defining a custom pad character to display during the prepare phase.
*/
byte padChar[8] = {
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111,
  B11111
};

/*
  Print a given time in milliseconds on the given row, prefixing it with the given string
*/
void printTime(int row, String s, long millis) {
  /*
     Transform a millisecond time rerresentation, like 1234567
     in a string representation like 20:34:56
  */
  int cent = millis % 1000;  // 1234567 -> 567
  millis -= cent;            // 1234567 -> 1234000
  cent /= 10;                // 567     -> 56 (no rounding, just floor)

  millis /= 1000;            // 1234000 -> 1234
  int second = millis % 60;  // 1234    -> 34
  millis -= second;          // 1234    -> 1200

  int minute = millis / 60;  // 1200    -> 20

  /*
     Pad the string to 8 chars, if needed
  */
  int pad = 8 - s.length();
  for (int i = 0; i < pad; i++) {
    s += " ";
  }

  /*
     Add minutes, seconds and cent, left padding with 0 if needed
  */
  if (minute < 10) s += "0"; s += minute; s += ":";
  if (second < 10) s += "0"; s += second; s += ":";
  if (cent < 10) s += "0"; s += cent;

  /*
     Pint the resulting string
  */
  printRow(row, s);
}

/*
    Print a given string on a given row
*/
void printRow(int row, String s) {
  lcd.setCursor(0, row);
  lcd.print(s);
}

/*
   Clean a row of the display, by printing a 16-blank string
*/
void clearRow(int row) {
  String s = "                ";
  lcd.setCursor(0, row);
  lcd.print(s);
}

/*
   Setup all the stuff
*/
void setup() {
  leftHandCapacitySensor.set_CS_AutocaL_Millis(0xFFFFFFFF);
  rightHandCapacitySensor.set_CS_AutocaL_Millis(0xFFFFFFFF);
  lcd.begin(16, 2);
  lcd.noCursor();
  lcd.display();
  lcd.createChar(0, padChar);
  Serial.begin(9600);
  versionString = "Firmware v.";
  versionString += FIRMWARE_VERSION;
}

/*
   Ok, let's go - main loop starts here
*/
void loop() {
  /*
     At each loop iteration, read the value of each capacity sensor
     and compare it to the threshold, to decide if the sensor is
     being touched.
  */
  boolean leftTouched  = leftHandCapacitySensor.capacitiveSensor(30)  > CAPACITY_SENSOR_THRESHOLD;
  boolean rightTouched = rightHandCapacitySensor.capacitiveSensor(30) > CAPACITY_SENSOR_THRESHOLD;

  switch (currentMode) {
    /*
       INIT mode: simply print the version and switch to IDLE mode
    */
    case MODE_INIT:
      printRow(0, "Stacking Chrono");
      printRow(1, versionString);
      currentMode = MODE_IDLE;
      break;

    /*
       IDLE mode: if both the sensor are being touched, switch to PREPARE mode,
       otherwise to nothing and wait
    */
    case MODE_IDLE:
      if (leftTouched && rightTouched) {
        clearRow(1);
        prepareTime = millis();
        prepareCursor = 0;
        currentMode = MODE_PREPARE;
      }
      break;

    /*
       PREPARE mode: if both the sensor are being touched long enough
       (that is, the row has been filled with 16 pad char),
       switch to READY mode, otherwise switch (back) to STOP
    */
    case MODE_PREPARE:
      if (leftTouched && rightTouched) {
        long now = millis();
        if (now  - prepareTime >= PREPARE_INTERVAL_SLOT) {
          prepareTime = now;
          lcd.setCursor(prepareCursor++, 1);
          lcd.write(byte(0));
        }
        if (prepareCursor == 16) {
          currentMode = MODE_READY;
        }
      } else {
        currentMode = MODE_STOP;
      }
      break;

    /*
       READY mode: as soon as one hand leaves the sensor, start counting time and switch to RUN mode
    */
    case MODE_READY:
      printRow(1, " *-* READY! *-* ");
      if (!(leftTouched && rightTouched)) {
        currentMode = MODE_RUNNING;
        startTime = millis();
      }
      break;

    /*
       RUNNING mode: calculate elapsed time and display it on the screen.
       As soon as both hands touch the sensors, stop the timer and go in STOP mode.
    */
    case MODE_RUNNING:
      if (leftTouched && rightTouched) {
        currentMode = MODE_STOP;
        lastTime = millis() - startTime;
      } else {
        printTime(1, "Elapsed", millis() - startTime);
      }
      break;

    /*
       STOP mode: display the last and best time, if any, then switch to IDLE mode
    */
    case MODE_STOP:
      startTime = 0;
      clearRow(1);

      /*
          Print last elapsed time if any and in case evaluate best time
      */
      if (lastTime != 0) {
        printTime(1, "Last:", lastTime);
        if (bestTime == 0 || bestTime > lastTime) {
          bestTime = lastTime;
        }
      }

      /*
         Print best time if any
      */
      if (bestTime != 0) {
        printTime(0, "Best:", bestTime);
      }

      currentMode = MODE_IDLE;
      delay(500);
      break;
  }
  /*
      That's it! Main loop ends here.
      Adding a 10 millis delay to improve reading of the sensors.
  */
  delay(10);
}

